package GlogJSONParser;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.json.*; // references .jar file

import com.lightysoft.logmx.business.ParsedEntry;
import com.lightysoft.logmx.mgr.LogFileParser;

/**
 * This parser takes JSONArray formatted logs using Glog keys to display in LogMX.
 * 
 * Please visit :
 * [http://high5-wiki/pages/viewpage.action?title=Glog+JSON+Format&spaceKey=GD]
 * for information on the Glog JSON Format used by High 5 Games.
 * 
 * @author Robert Mitola, LogMX Team
 */
public class GlogJSONParser extends LogFileParser {
	
	// StringBuilder to construct the entire JSON entry to be parsed
	private StringBuilder entryBuilder = new StringBuilder();
	// String to hold the entire JSON entry after it is constructed by parseLine
	private String entryString;
	// the entry
	private ParsedEntry entry;
	// the JSON array to hold all incoming logs
	JSONArray jArray;
	// the JSON object to store each element of the jArray
	JSONObject jObj;
	// user-defined fields
	private static final String PROG_TIMESTAMP_FIELD_KEY = "Received at...";
	private static final String THROWABLE_FIELD_KEY =  "throwable";
	private static final String EXTRA_FIELD_KEY = "extra";
	//private static final List<String> EXTRA_FIELDS_KEYS = Arrays
	        //.asList(PROG_TIMESTAMP_FIELD_KEY, THROWABLE_FIELD_KEY, EXTRA_FIELD_KEY);
	private static final List<String> EXTRA_FIELDS_KEYS = Arrays
	        .asList();
            
	// date formatting
    //http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html#SimpleDateFormat()
	DateFormat df = new SimpleDateFormat("MM/dd HH:mm:ss:SSS");
	/*
	 * number of "s in the pLine - used to determine if JSON formatting characters
	 * exist inside or outside of the JSON Object's key names / key values or not -
	 * this prevents format parsing of characters not meant to be for formatting
	*/
	private int quoteCount = 0;
	
	/**
	 * Returns the absolute date of entry.
	 * 
	 * @param pEntry the parsed entry to retrieve the absolute date from
	 * @return the Date of the entry
	 * @see com.lightysoft.logmx.mgr.LogFileParser#getAbsoluteEntryDate(com.lightysoft.logmx.business.ParsedEntry)
	 * @throws Exception
	 */
	@Override
	public Date getAbsoluteEntryDate(ParsedEntry pEntry) throws Exception {
		// TODO Auto-generated method stub
		return null;
	}

	/** 
     * Returns the name of this parser.
     * 
     * @return a String denoting the name of this parser
     * @see com.lightysoft.logmx.mgr.LogFileParser#getParserName()
     */
	@Override
	public String getParserName() {
		return "H5G-Glog JSON Parser";
	}

	/**
     * Returns a relative Date for the given entry (if entry's ExtraInfo contains "1265", 
     * it means "T0 + 1265 ms", so simply return "new Date(1265)").
     *
     * @see com.lightysoft.logmx.mgr.LogFileParser#getRelativeEntryDate(com.lightysoft.logmx.business.ParsedEntry)
     * @throws Exception
     */
	@Override
	public Date getRelativeEntryDate(ParsedEntry pEntry) throws Exception {
		// TODO Auto-generated method stub
		return null;
	}

	/**
     * Returns the supported file type for this parser.
     * 
     * @return a String denoting the type of log files this parser will be able to parse
     * @see com.lightysoft.logmx.mgr.LogFileParser#getSupportedFileType()
     */
	@Override
	public String getSupportedFileType() {
		return "H5G-Glog JSON Log Files";
	}
	
	/** 
     * Returns the ordered list of user-defined fields to display (given by their key), for each entry.
     * 
     * @return a List<String> containing all user-defined fields to log
     * @see com.lightysoft.logmx.mgr.LogFileParser#getUserDefinedFields()
     */
    @Override
    public List<String> getUserDefinedFields() {
    	return EXTRA_FIELDS_KEYS;
    }

    /**
     * Process the new line of text read from file.
     * 
     * @param pLine the String to parse
     * @see com.lightysoft.logmx.mgr.LogFileParser#parseLine(java.lang.String)
     * @throws Exception
     */
	@Override
	protected void parseLine(String pLine) throws Exception {
		
		// if there is no new content, record the previous entry (is exists)
		if (pLine == null) {
            recordPreviousEntryIfExists();
            return;
        }
		
		// handle new JSON array content or possible newline characters in logs
		if(entryBuilder.length() == 0){ // if this is a new JSON
			// trim the \n at beginning of JSON String, if this pLine is the beginning
			pLine = (pLine.indexOf('[') > -1) ? pLine.substring(pLine.indexOf('[')) : pLine; 
		}else if((quoteCount & 1) != 0){
			pLine = "\\n" + pLine; // else there is a newline char in a log field/value
		}

		String excludeFakeQuotes = pLine.replace("\\\"", "xx"); // prevents quotation marks that are part of JSON fields from being counted
		quoteCount += pLine.length() - excludeFakeQuotes.replace("\"", "").length(); // number of quotation marks
		
		// handle complete or incomplete reading of JSON array
		if((quoteCount & 1) == 0 // even number of quotation marks?
				&& pLine.length() > 0 // not an empty line?
				&& pLine.charAt(pLine.length() - 1) == ']'){ // end of JSON character?
			entryBuilder.append(pLine);
			quoteCount = 0; // reset quoteCount
			prepareNewEntry();
			setEntry(); // *
		}else{
			// just append the new pLine without setting any of the entry
			entryBuilder.append(pLine);
		}
		
	}
	
	/**
	 * Records a completed JSONArray entry and parses it using json-20141113.jar
	 * functionality.
	 * 
	 * @throws Exception
	 */
	private void setEntry(){
		
		entryString = entryBuilder.toString();
		entryBuilder = new StringBuilder();
		Iterator<String> keyIterator; // to iterate through all user-defined fields
		String key; // current user-defined field
		Date time; // for timestamp conversion
		jArray = new JSONArray(entryString);
		for(int i = 0; i < jArray.length(); i++){
			try{
				// flush the entry
				entry = createNewEntry();
				jObj = jArray.getJSONObject(i);
				entry.setUserDefinedFields(new HashMap<String, Object>(EXTRA_FIELDS_KEYS.size())); // new extra fields method
				// SET ENTRY DETAILS //
				// emitter - appends id to beginning for multiple source differentiation
				if(jObj.has("logger")){
					String emitter = jObj.getString("logger");
					if(jObj.has("id")){
						emitter = jObj.getString("id") + "." + emitter;
					}
					entry.setEmitter(emitter);
				}
				// timestamp - determined by log
				if(jObj.has("timestamp")){
					if(jObj.get("timestamp") instanceof String){ // timestamp sent as a String?
						time = new Date(Long.parseLong(jObj.getString("timestamp")));
					}else{ // timestamp was sent as something other than a String
						time = new Date((Long) jObj.get("timestamp"));
					}
					entry.setDate(df.format(time));
				}
				// level - the level of importance
				if(jObj.has("level"))
					entry.setLevel(jObj.getString("level")); // java.net.URLDecoder.decode(text, "UTF-8") 	[Decoder if we need to URL-decode again]
				// thread - the thread the log was sent from
				if(jObj.has("thread"))
					entry.setThread(jObj.getString("thread"));
				// message - details about this entry
				if(jObj.has("message"))
					entry.setMessage(jObj.getString("message"));
					//entry.setMessage(jObj.getString("message"));
				// extra fields - as defined by the user/programmer
				keyIterator = EXTRA_FIELDS_KEYS.iterator();
				while(keyIterator.hasNext()){
					key = keyIterator.next();
					if(jObj.has(key))
						entry.getUserDefinedFields().put(key, (jObj.getString(key)));
				}
				// received at... field - when this log was parsed
				entry.getUserDefinedFields().put(PROG_TIMESTAMP_FIELD_KEY, df.format(new Date()));
				
				// add the entry
				addEntry(entry);
			
			}catch(Exception e){
				e.printStackTrace();
			}
		}	
		
	}
	
	/**
	 * Prepares a new log entry by first recording the previous if it exists and
	 * resetting the ParsedEntry object used by this parser.
	 * 
	 * @throws Exception
	 */
	private void prepareNewEntry() throws Exception {
        recordPreviousEntryIfExists();
        entry = createNewEntry();
    }
	
	/**
	 * Adds a log entry if the ParsedEntry object is not null (i.e. has log content).
	 * 
	 * @throws Exception
	 */
	private void recordPreviousEntryIfExists() throws Exception {
        if (entry != null) {
            addEntry(entry);
        }
    }
	
}
