package GlogManager;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.Queue;

import javax.swing.Icon;

import com.lightysoft.logmx.business.LogURL;
import com.lightysoft.logmx.business.LogURLParameter;
import com.lightysoft.logmx.business.LogURLParameterType;
import com.lightysoft.logmx.mgr.LogFileInfo;
import com.lightysoft.logmx.mgr.LogStreamManager;
import com.lightysoft.logmx.mgr.ManagerSpecVersion;

/**
 * This manager receives JSONArray formatted Strings with Glog details on a 
 * specified port.
 * 
 * Please visit :
 * [http://high5-wiki/pages/viewpage.action?title=Glog+JSON+Format&spaceKey=GD]
 * for information on the Glog JSON Format used by High 5 Games.
 * 
 * @author Robert Mitola, LogMX Team
 */
public class GlogSocketManager extends LogStreamManager {

	private String url = null;
	private int port = 0;
	private static LogURL templateLogURL = null;
	public static final String PROTOCOL = "glog";
    private static final Icon FILE_TYPE_ICON = getIconFile("gfx_live_stream.png");
    public static final Icon FILE_TYPE_ICON_32 = getIconFile("gfx_live_stream32.png");
    private static final String URL_PARAM_PORT = "Port";
    private static final String URL_PARAM_PROTOCOL = "Protcol";
    private static final String URL_PARAM_IP = "Interface IP";
    private static final String URL_PATTERN = PROTOCOL + "://{" + URL_PARAM_PROTOCOL
            + "}:{" + URL_PARAM_IP + "}:{" + URL_PARAM_PORT + "}";
    public static final String UDP_PROTOCOL = "TCP with UDP Auto-Discovery";
    public static final String TCP_PROTOCOL = "TCP";
    private SocketManagerServer sms;
    // lock used to prevent the manager from reading lines before any are sent
    private Object glogLock;
    // boolean to prevent the manager from reading lines before any are sent
    public boolean waitBool = true;
    // queue to hold lines sent from sources to read in
    private Queue<String> toRead;
    // for TCP or UDP connections
    private boolean isTCP = true;
    // interface IP for incoming connections/packets
    public String ip;
    
    /**
     * Sets read timeout for the SocketManagerServer.
     * 
     * @param arg0 determines if read timeout is used or not
     */
	@Override
	public void setHasReadTimeout(boolean arg0) {
		// I had no use for this while developing the manager, but this can be used to set a read-timeout for the SocketManagerServer
	}

	/**
     * Called by LogMX to know when the opened resource was modified for the last time.<BR/>
     * The expected result is a standard Java timestamp (number of milliseconds since 01/01/1970, 00:00:00 GMT). 
     * For example, if you have a Java Date, you can get such a timestamp using {@link java.util.Date#getTime()}.<BR/>
     * <BR/>
     * <B>This method may be called periodically, so it should be rather fast (e.g. no socket connection, but fast I/O operations).</B><BR/>
     * <BR/>
     * If the Manager is not able to get this timestamp, it may return 0, but file-change detection will be affected.
     *  
     * @return
     *      A Java-based timestamp of last modification, or 0 if Manager can't get this timestamp
     */
	@Override
	public long getCurrentFileLastModifDate() {
		return 0; // I left this at returning 0 since the manager is reading a socket, not a file
	}

	/**
     * Called by LogMX to get the current opened URL, matching this Manager URL template.<BR/>
     * This method may return the URL string received by {@link com.lightysoft.logmx.mgr.LogFileManager#init(LogURL, String)}, 
     * or use {@link com.lightysoft.logmx.mgr.LogFileManager#getURLFromLogURL(LogURL, String)} to construct the URL. 
     * 
     * @return
     *      Current opened URL
     */
	@Override
	public String getCurrentURL() {
		return url;
	}

	/**
     * Called by LogMX to get this Manager icon, which must be a 16x16 icon.<BR/> 
     * Manager may use {@link com.lightysoft.logmx.mgr.LogFileManager#getIconFile(String)} to get an icon file
     * contained in LogMX pictures directory, in order to avoid relative file path issues.<BR/> 
     * This icon may (should) be cached, the same instance may be returned for each call for better performances.<BR/>
     * If <code>null</code> is returned, a default icon will be used.
     *  
     * @return
     *      Manager icon, or <code>null</code> to use a default icon.
     */
	@Override
	public Icon getFileTypeIcon() {
		return FILE_TYPE_ICON;
	}
	
	/**
	 * See getFileTypeIcon(), but uses a 32x32 icon
	 */
	@Override
    public Icon getFileTypeIcon32x32() {
        return FILE_TYPE_ICON_32;
    }

	/**
     * Called by LogMX to get the name of this Manager.
     * This name can contain any character and is displayed in LogMX GUI.
     * 
     * @return 
     *      Manager name
     */
	@Override
	public String getName() {
		return "H5G-Glog Socket Manager";
	}

	/**
     * Called by LogMX to get the Protocol handled by this Manager.<BR>
     * This protocol can be any string, like "http", "jdbc", "ftp", "myprotocol", ...<BR>
     * LogMX will use it to find the Manager to use for a given a URL (eg: "glog://4444") 
     * 
     * @return
     *      Protocol handled by this Manager
     */
	@Override
	public String getProtocolName() {
		return PROTOCOL;
	}

	/**
     * Called by LogMX to get the Manager Specification version supported by this Manager.
     * For example, this method can return <code>ManagerSpecVersion.V1_2</code>
     *   
     * @return
     *      Supported Manager Specification version
     */
	@Override
	public ManagerSpecVersion getSpecificationVersion() {
		return ManagerSpecVersion.V1_2;
	}

	/**
     * Called by LogMX to get a template of <code>LogURL</code> used by this Manager.<BR/>
     * 
     * <BR/>
     * The returned <code>LogURL</code> can (should) be a static shared instance for better performance.
     * LogMX will not use it directly but will use a clone of it. 
     * 
     * @return 
     *      <code>LogURL</code> used by this Manager
     */
	@Override
	public LogURL getTemplateLogURL() {
		if (templateLogURL == null) {
            templateLogURL = new LogURL();
            templateLogURL.addParameter(new LogURLParameter(URL_PARAM_PORT,
                    LogURLParameterType.INTEGER));
            templateLogURL.addParameter(new LogURLParameter(URL_PARAM_PROTOCOL,
                    LogURLParameterType.ENUM, TCP_PROTOCOL, false, new Object[] { TCP_PROTOCOL,
                    UDP_PROTOCOL }));
            try {
				templateLogURL.addParameter(new LogURLParameter(URL_PARAM_IP,
						LogURLParameterType.STRING, InetAddress.getLocalHost().toString()));
			} catch (UnknownHostException e) {
				templateLogURL.addParameter(new LogURLParameter(URL_PARAM_IP,
						LogURLParameterType.STRING)); // leave default blank
				System.out.println("Error getting local host IP for default IP templateLogURL.");
				e.printStackTrace();
			}
		}
		return templateLogURL;
	}

	/**
     * Called by LogMX to know how to display a URL handled by this Manager.
     * 
     * @return URL Pattern used by this Manager
     */
	@Override
	public String getURLPattern() {
		return URL_PATTERN;
	}

	/**
     * Called by LogMX to initialize the Manager for the specified resource.<BR/>
     * This method shouldn't try to open the resource yet, but should only save URL parameters 
     * for further use.
     * 
     * @param pLogURL
     *      LogURL for the resource this Manager will have to open
     * @param pURLString
     *      String representation for this URL (may be helpful for the Manager, especially to implement 
     *      {@link com.lightysoft.logmx.mgr.LogFileManager#getCurrentURL()}
     * @throws Exception
     *      If the Manager is not able to process this URL
     */
	@Override
	public void init(LogURL pLogURL, String pURLString) throws Exception {
		LogURLParameter portParam = pLogURL.getParameter(URL_PARAM_PORT);
		LogURLParameter protParam = pLogURL.getParameter(URL_PARAM_PROTOCOL);
		LogURLParameter ipParam = pLogURL.getParameter(URL_PARAM_IP);
		port = (Integer) portParam.getValue();
		String protocol = (protParam.getValue() == null) ? TCP_PROTOCOL : protParam.getValue()
	            .toString();
		isTCP = protocol.equals(TCP_PROTOCOL);
		ip = ipParam.getValue().toString();
		url = pURLString;
		if (!isTCP && !protocol.equals(UDP_PROTOCOL)) {
            throw new Exception("Invalid protocol \"" + protocol + "\"");
        }
	}

	/**
     * Open the resource specified by {@link com.lightysoft.logmx.mgr.LogFileManager#init(LogURL, String)} to get ready 
     * to read, and return information on this resource (file size and date).<BR/>
     * If the Manager is not able to get the number of bytes to read, it can return a negative number.
     * As this number is used to display a progress bar while the Manager is loading the resource, returning a
     * negative number will make the progress bar stays at "100%" all the time.
     *  
     * @return
     *          File information structure (containing file size/date).
     * @throws Exception
     *      If the resource couldn't be opened (this Exception message will be displayed in LogMX GUI)
     */
	@Override
	public LogFileInfo prepareForReading() throws Exception {
		// set toRead so that it can start receiving logs
		toRead = new LinkedList<String>();
		// set glogLock so it can be used with the threads
		glogLock = new Object();
		// instantiate the socket manager server (as a new thread)
		sms = new SocketManagerServer(port, isTCP, glogLock, this);
		// start the server
		sms.start();
		return new LogFileInfo(-1, 0);
	}

	/**
     * Called by LogMX to get the next line of text from the underlying Manager resource (file, stream, socket,...).<BR/>
     * Manager must return <B><code>null</code></B> if there is no more byte to read.
     * 
     * @return
     *      Next line, or <B><code>null</code></B> if there is no more byte to read.
     * @throws Exception
     *      If an error occurred while reading bytes. Managers should not handle Exceptions but should
     *      throw them instead, so that LogMX can catch them.
     */
	@Override
	public String readLine() throws InterruptedException {
		
		// wait for first log to come through before trying to read anything
		if(waitBool == true){
			synchronized(glogLock){
				glogLock.wait(); // glogLock.notify() can be found in SocketManagerServerThread
			}
			waitBool = false; // only need to wait for first log - no more than that
		}
		
		if(sms == null) // if the socket manager server is null
			return null;
		
		// return the next String in the toRead queue
		synchronized(READ_MUTEX){
			return toRead.poll();
		}
		
	}

	/**
     * Called by LogMX when it doesn't need to use this Manager anymore.<BR/>
     * This method should release all resources allocated by this Manager (e.g. close file, socket,...)<BR/>
     * Once this method is called, LogMX will not call <code>readLine()</code>.<BR/>
     */
	@Override
	public void releaseResources(boolean pSoftRelease) {
		if (!pSoftRelease) {
			if (sms != null) {
				try {
					sms.releaseResources();
				} catch (IOException e) {
					System.out.println("Could not close server socket at port: " + port);
					e.printStackTrace();
				}
            }
		}
	}
	
	/**
	 *  Used to display file name in GUI file tab (when Manager is not ready yet),
	 *  without the trailing "?s=1", if any.
	 *  
	 * @param pFileName the file name
	 * @return the shortened version of the file name
	 */
    public static String getShortFileName(String pFileName) {
        return "GLOG";
    }

    /**
     * Adds a String to the toRead queue, which receives logs from multiple sources
     * and passes them to readLine() one at a time.
     * 
     * @param s String to add
     */
    public void addToRead(String s){
    	toRead.add(s);
    }

}
