package GlogManager;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.ServerSocket;

/**
 * This is the SocketManagerServer used with GlogSocketManager to create new threads 
 * associated with the same socket, effectively allowing multiple log source clients
 * to connect to the GlogSocketManager at once.
 * 
 * Please visit :
 * [http://high5-wiki/pages/viewpage.action?title=Glog+JSON+Format&spaceKey=GD]
 * for information on the Glog JSON Format used by High 5 Games.
 * 
 * @author Robert Mitola
 */
public class SocketManagerServer extends Thread {

	// id for each socket thread (deprecated)
	// this is now deprecated, but I am leaving it in for now, in case it
	// becomes of use to us
	private static int id = 0;
	private int portNumber = 0; // port number receiving logs on
	private Object glogLock; // the glogLock passed from GlogSocketManager
	private ServerSocket serverSocket; // the TCP server socket this server will control
	private GlogSocketManager owner; // the GlogSocketManager that instantiated this server
	private boolean listening = true; // if the server socket should listen for connections
	private boolean isTCP = true; // TCP or UDP if false
	private static final String DISCOVERY_JSON = "[{\"action\":\"auto-discovery-signal\"}]"; // UDP discovery to establish TCP
	public static final String POLICY_XML = // policy XML to send to Flash log sources
			"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
			"<cross-domain-policy>" +
	  			"<site-control permitted-cross-domain-policies=\"master-only\"/>" +
	  			"<allow-access-from domain=\"*\" to-ports=\"*\"/>" +
	  		"</cross-domain-policy>\u0000";
	private String informationJSON; // information String to send to connection seeking UDP packets
	
	// UDP Specific Stuff ////////////////////////////////////
	
    private DatagramSocket datagramSocket = null; // the UDP datagram socket this server will control
    private boolean morePackets = true; // boolean to let UDP server know to keep accepting packets or not
	
	//////////////////////////////////////////////////////////
	
	/**
	 * Constructor.
	 * 
	 * @param portNumber the port number for the server socket to listen on
	 * @param lock the glogLock passed from GlogSocketManager
	 * @throws IOException
	 */
	public SocketManagerServer(int portNumber, boolean isTCP, Object lock, GlogSocketManager owner) throws IOException {
		super("SocketManagerServer");
		this.portNumber = portNumber; // port number receiving logs on
		this.glogLock = lock; // the glogLock passed from GlogSocketManager
		this.owner = owner; // the GlogSocketManager that instantiated this server
		this.isTCP = isTCP; // TCP or UDP if false
		informationJSON = "[{"
				+ "\"action\":\"identity-discovered\","
				+ "\"ip\":\"" + owner.ip + "\","
				+ "\"port\":\"" + this.portNumber + "\""
				+ "}]";
		if(isTCP)
			this.serverSocket = new ServerSocket(portNumber); // the TCP server socket this server will control
		else{ // UDP
			this.serverSocket = new ServerSocket(portNumber); // possible TCP after UDP finding
			this.datagramSocket = new DatagramSocket(portNumber); // the UDP datagram socket this server will control
		}
	}
	
	/**
	 * Runs the SocketManagerServer thread, accepting all connections on the
	 * specified port.
	 */
	public void run(){
	    
		if(isTCP){
		// TCP Server ///////////////////////////////////////////////////////////////////
		
			if (portNumber < 0) {
				System.err.println("Usage: java SocketManagerServer <port number>");
			}
			// Loop while listening to establish connections to all sending sources    
			try{ 
				while (listening) {	
					// id is deprecated... considering removing, but leaving in for now
					new SocketManagerServerThread(serverSocket.accept(), 
							id++, glogLock, owner).start();
				}
	    	}catch(IOException e){
	            System.err.println("Could not listen on port " + portNumber);
	        }
	    
        /////////////////////////////////////////////////////////////////////////////////
		}else{
	    // UDP Server ///////////////////////////////////////////////////////////////////
			
			// default packet
			byte[] buf = new byte[256];
			DatagramPacket packet = new DatagramPacket(buf, buf.length);
			
			while(morePackets){
				try{
					// receive packet
					datagramSocket.receive(packet);
					// get the message from the packet
					String received = new String(packet.getData(), 0, packet.getLength());
					System.out.println("UDP Received:" + received);
					// If this line is the Flash Policy File Request
					if(received.contains("<policy-file-request")){
						// get received packet's information
						InetAddress address = packet.getAddress();
		                int port = packet.getPort();
		                // set packet's message to the policy file
		                buf = POLICY_XML.getBytes();
		                packet = new DatagramPacket(buf, buf.length, address, port);
		                // send the policy file
		                datagramSocket.send(packet);
					}else if(received.equals(DISCOVERY_JSON)){ // UDP discovery to establish TCP
						// set isTCP to true to establish TCP server
						isTCP = true;
						// get received packet's information
						InetAddress address = packet.getAddress();
		                int port = packet.getPort();
		                // set packet's message to the policy file
		                buf = informationJSON.getBytes();
		                packet = new DatagramPacket(buf, buf.length, address, port);
		                // send the information file
		                datagramSocket.send(packet);
		                // stop looping here
		                morePackets = false;
		                // close the datagram (UDP) socket
		                datagramSocket.close();
		                // run again, now with TCP
		                run();
					}else{ // otherwise it is a log
            			// adds this line to the queue of lines to manage
            			owner.addToRead(received);
            			if(owner.waitBool == true){
            				synchronized(glogLock){
            					glogLock.notify(); // notifies the GlogSocketManager lock
            					// to unlock, as text is not being transmitted at least once
            				}
            			}
					}
				}catch(IOException e){
					e.printStackTrace();
	                morePackets = false;
				}
			}
	    
	    /////////////////////////////////////////////////////////////////////////////////
		}
	}
	
	/**
	 * Tells this server to stop listening on its server socket's port
	 * and closes said server socket.
	 * 
	 * @throws IOException
	 */
	public void releaseResources() throws IOException{
		listening = false;
		morePackets = false;
		if(isTCP)
			serverSocket.close(); // TCP
		else{
			serverSocket.close(); // TCP
			datagramSocket.close(); // UDP
		}
		System.out.println("server's socket closed");
	}
	
}
