package GlogManager;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;

/**
 * This is a SocketManagerServerThread created by GlogSocketManagerServer
 * as a new TCP connection's independent thread. Each connecting source has its
 * own instantiation of this, and calls GlogSocketManager.addToRead(inputLine) 
 * to pass text received from its source to the manager.
 * 
 * Please visit :
 * [http://high5-wiki/pages/viewpage.action?title=Glog+JSON+Format&spaceKey=GD]
 * for information on the Glog JSON Format used by High 5 Games.
 * 
 * @author Robert Mitola
 */
public class SocketManagerServerThread extends Thread {

	private Socket socket = null; // the socket to receive messages on
	private Object glogLock; // the glogLock passed from SocketManagerServer
	private GlogSocketManager owner; // the GlogSocketManager that this thread sends logs to
	private int id; // id for this socket thread (deprecated)

	/**
	 * Constructor.
	 * 
	 * @param socket the socket to receive messages on
	 * @param id id for each socket thread (deprecated)
	 * @param glogLock the glogLock passed from SocketManagerServer
	 */
    public SocketManagerServerThread(Socket socket, int id, Object glogLock, GlogSocketManager owner) {
        super("SocketManagerServerThread");
        this.socket = socket; // the socket to receive messages on
        this.glogLock = glogLock; // the glogLock passed from SocketManagerServer
        this.owner = owner; // the GlogSocketManager that this thread sends logs to
        this.id = id; // id for this socket thread (deprecated)
    }
    
    /**
	 * Runs this SocketManagerServerThread thread, passing received
	 * logs to the GlogSocketManager.
	 */
    public void run() {

        try{
        	BufferedReader in = new BufferedReader(
    				new InputStreamReader(socket.getInputStream(), "UTF-8"));
        	
        	// Send policy file to log source
        	sendPolicyFile();
        	System.out.println("New Thread: " + id);
        	
            // the text received from this thread's source
        	String inputLine;
        	// while the source is connected...
            while ((inputLine = in.readLine()) != null) {
            	
            	// If this line is the Flash Policy File Request
            	if(inputLine.contains("<policy-file-request")){
            		/*
            		The policy XML has to be sent before the request is received,
            		presumably because of network timing. This if block now serves
            		to prevent this SocketManagerServerThread from trying to
            		read in the policy file request.
            		*/
            		System.out.println("Flash Policy Request Received");
            	}
            	// make sure this line is not just the null character,
            	// otherwise parsing it will throw an error
            	else if(!(inputLine.length() == 1 && 
            		(Character.getNumericValue(inputLine.charAt(0))) == -1)){
            		
            		//System.out.println(id + ":" + (++owner.num) + ": " + inputLine);
            		 
            		// adds this line to the queue of lines to manage
            		owner.addToRead(inputLine);
            		if(owner.waitBool == true){
            			synchronized(glogLock){
            				glogLock.notify(); // notifies the GlogSocketManager lock
            				// to unlock, as text is not being transmitted at least once
            			}
            		}
            	}
            }
            socket.close(); // close the socket when the connection is gone
            in.close(); // close the buffered reader
            System.out.println("Socket closed");
        }catch (IOException e){
            e.printStackTrace();
        }
    }
    
    /**
     * Sends the Flash policy XML String to the log source.
     * 
     * @throws IOException
     */
    private void sendPolicyFile() throws IOException{
       	PrintWriter out = new PrintWriter(socket.getOutputStream());
       	out.write(SocketManagerServer.POLICY_XML);
       	out.flush();
    }

}
